import time
import threading
import logging
from ...geonet.service_access_point import Area, GeoBroadcastHST, HeaderType
from .denm_coder import DENMCoder
from ...btp.router import Router as BTPRouter
from ...btp.service_access_point import BTPDataRequest, CommonNH, CommunicationProfile
from ...applications.road_hazard_signalling_service.service_access_point import (
    DENRequest,
)


class VehicleData:
    """
    Class that stores the vehicle data.

    Attributes
    ----------
    station_id : int
        Station ID as specified in ETSI TS 103 831 V2.1.1 (2022-11).
    station_type : int
        Station Type as specified in ETSI TS 103 831 V2.1.1 (2022-11).
    """

    def __init__(self) -> None:
        self.station_id = 0
        self.station_type = 0


class DecentralizedEnvironmentalNotificationMessage:
    """
    Decentralized Environmental Notification Message class.

    Attributes
    ----------
    denm : dict
        All the DENM message in dict format as decoded by the DENMCoder.
    sequence_number : int
        Sequence number to serialize the DENMs.
    """

    def __init__(self) -> None:
        self.denm = self.generate_white_denm()
        self.sequence_number = 0

    def generate_white_denm(self) -> dict:
        """
        Generate a white DENM.
        """
        white_denm = {
            "header": {"protocolVersion": 2, "messageId": 1, "stationId": 0},
            "denm": {
                "management": {
                    "actionId": {
                        "originatingStationId": 0,  # 4294967295
                        "sequenceNumber": 0,  # 65535
                    },
                    "detectionTime": 0,
                    "referenceTime": 0,
                    "termination": "isCancellation",
                    "eventPosition": {
                        "latitude": 900000001,
                        "longitude": 1800000001,
                        "positionConfidenceEllipse": {
                            "semiMajorConfidence": 4095,
                            "semiMinorConfidence": 4095,
                            "semiMajorOrientation": 3601,
                        },
                        "altitude": {
                            "altitudeValue": 800001,
                            "altitudeConfidence": "unavailable",
                        },
                    },
                    "relevanceDistance": "lessThan50m",
                    "relevanceTrafficDirection": "allTrafficDirections",
                    "validityDuration": 0,
                    "TransmissionInterval": 100,
                    "stationType": 0,
                }
            },
        }
        return white_denm

    def fullfill_with_denrequest(self, request: DENRequest) -> None:
        """
        Fullfill the DENM with the DENRequest data for the Road Hazard Signalling aplication.
        """
        # Add Management Container
        self.denm["denm"]["management"]["detectionTime"] = request.detection_time
        self.denm["denm"]["management"]["referenceTime"] = int(
            (time.time() - 1072911600) * 1000
        )
        self.denm["denm"]["management"]["TransmissionInterval"] = request.denm_interval

        self.denm["denm"]["management"][
            "relevanceDistance"
        ] = request.relevance_distance
        self.denm["denm"]["management"][
            "relevanceTrafficDirection"
        ] = request.relevance_traffic_direction
        # self.denm['denm']['management']['termination'] = request.DENMTermination
        # --> TODO: [OPTIONAL] to be implemented

        self.denm["denm"]["management"][
            "eventPosition"
        ] = request.event_position  # to be defined
        self.denm["denm"]["management"]["stationType"] = request.rhs_vehicle_type

        # Add Situation Container
        situation_container = {
            "informationQuality": request.quality,
            "eventType": {
                "ccAndScc": (request.rhs_cause_code, request.rhs_subcause_code)
            },
        }
        self.denm["denm"]["situation"] = situation_container

        # Add Location Container
        location_container = {
            "eventSpeed": {
                "speedValue": request.rhs_event_speed,
                "speedConfidence": int(request.confidence / 2),
            },
            "eventPositionHeading": {
                "value": request.heading,
                "confidence": request.confidence,
            },
            "detectionZonesToEventPosition": [
                [
                    {
                        "pathPosition": {
                            "deltaLatitude": 131072,
                            "deltaLongitude": 131072,
                            "deltaAltitude": 12800,
                        }
                    }
                ]
            ],
        }
        self.denm["denm"]["location"] = location_container

    def fullfill_with_collision_risk_warning(self, request: DENRequest) -> None:
        """
        Fullfill the DENM with the DENRequest data for the Collision Risk Warning aplication.
        """
        # Add Management Container
        self.denm["denm"]["management"]["detectionTime"] = request.detection_time
        self.denm["denm"]["management"]["referenceTime"] = int(
            (time.time() - 1072911600) % 65536
        )
        self.denm["denm"]["management"]["TransmissionInterval"] = request.denm_interval

        self.denm["denm"]["management"]["eventPosition"] = request.event_position

        # Add Situation Container
        situation_container = {
            "informationQuality": request.quality,
            "eventType": {
                "ccAndScc": (request.lcrw_cause_code, request.lcrw_subcause_code)
            },
        }
        self.denm["denm"]["situation"] = situation_container

    def fullfill_with_vehicle_data(self, vehicle_data: VehicleData) -> None:
        """
        Fullfill the DENM with vehicle data.
        """
        self.denm["header"]["stationId"] = vehicle_data.station_id
        self.denm["denm"]["management"]["stationType"] = vehicle_data.station_type
        self.denm["denm"]["management"]["actionId"][
            "originatingStationId"
        ] = vehicle_data.station_id
        self.denm["denm"]["management"]["actionId"][
            "sequenceNumber"
        ] = self.sequence_number
        self.sequence_number = (self.sequence_number + 1) % 65535


class DENMTransmissionManagement:
    """
    DENM Transmission Management class.
    This sub-function implements the protocol operation of the originating ITS-S, as specified in
    ETSI TS 103 831 V2.1.1 (2022-11) Section 8, including in particular:
    - Trigger a thread to start the generation of consecutive DENM messages.
    - The general parameters of the DENM are defined in the application layer.
    - Transmission of the DENMs.

    Attributes
    ----------
    btp_router : BTPRouter
        BTP Router object.
    vehicle_data : VehicleData
        Vehicle data object.
    denm_coder : DENMCoder
        DENM Coder object.
    crw_denm : DecentralizedEnvironmentalNotificationMessage
        DENM message with the Collision Risk Warning data.
    new_denm : DecentralizedEnvironmentalNotificationMessage
        New DENM message to be sent, as Emergency Vehicle Warning message.
    """

    def __init__(
        self, btp_router: BTPRouter, denm_coder: DENMCoder, vehicle_data: VehicleData
    ) -> None:
        """
        Initialize the DENM Transmission Management.
        """
        self.logging = logging.getLogger("denm_service")
        self.btp_router: BTPRouter = btp_router
        self.vehicle_data = vehicle_data
        self.denm_coder = denm_coder
        self.crw_denm = DecentralizedEnvironmentalNotificationMessage()
        self.new_denm = DecentralizedEnvironmentalNotificationMessage()

    def request_denm_sending(self, denm_request: DENRequest) -> None:
        """
        Request to send a DENM and starts a thread.
        """
        t = threading.Thread(target=self.trigger_denm_messages, args=[denm_request])
        t.start()

    def send_collision_risk_warning_denm(self, denm_request: DENRequest) -> None:
        """
        Request to send a single DENM message with the Collision Risk Warning data.
        """
        self.crw_denm = DecentralizedEnvironmentalNotificationMessage()
        self.crw_denm.fullfill_with_vehicle_data(self.vehicle_data)
        self.crw_denm.fullfill_with_collision_risk_warning(denm_request)
        self.transmit_denm(self.crw_denm)

    def trigger_denm_messages(self, denm_request: DENRequest) -> None:
        """
        Function to transmits consecutive DENM message.

        Parameters
        ----------
        denm_request : DENRequest
            DENM Request object.
        """
        transmission_time = 0

        while transmission_time < denm_request.time_period:
            self.new_denm = DecentralizedEnvironmentalNotificationMessage()
            self.new_denm.fullfill_with_vehicle_data(self.vehicle_data)
            self.new_denm.fullfill_with_denrequest(denm_request)

            self.transmit_denm(self.new_denm)
            time.sleep(denm_request.denm_interval / 1000)
            transmission_time += denm_request.denm_interval

    def transmit_denm(
        self, denm_to_send: DecentralizedEnvironmentalNotificationMessage
    ) -> None:
        """
        Function to fullfill and send a single DENM message.

        Parameters
        ----------
        denm_to_send : DecentralizedEnvironmentalNotificationMessage
            New DENM message to be sent.
        """
        request = BTPDataRequest()
        request.btp_type = CommonNH.BTP_B
        request.destination_port = 2002
        request.gn_packet_transport_type.header_type = HeaderType.GEOBROADCAST
        request.gn_packet_transport_type.header_subtype = (
            GeoBroadcastHST.GEOBROADCAST_CIRCLE
        )
        request.gn_area = Area()
        request.gn_area.a = 100
        request.gn_area.b = 0
        request.gn_area.angle = 0
        request.gn_area.latitude = denm_to_send.denm["denm"]["management"][
            "eventPosition"
        ]["latitude"]
        request.gn_area.longitude = denm_to_send.denm["denm"]["management"][
            "eventPosition"
        ]["longitude"]
        request.communication_profile = CommunicationProfile.UNSPECIFIED
        request.data = self.denm_coder.encode(denm_to_send.denm)
        request.length = len(request.data)
        self.btp_router.btp_data_request(request)
        self.logging.debug(
            "Recieved DENM with timestamp: %s, station_id: %s",
            denm_to_send.denm["denm"]["management"]["referenceTime"],
            denm_to_send.denm["header"]["stationId"],
        )



from .service_access_point import BTPDataRequest


class BTPAHeader:
    """
    BTP-A Header class.

    Specified in ETSI EN 302 636-5-1 V2.1.0 (2017-05). Section 7.2

    Attributes
    ----------
    destination_port : int
        (16 bit integer) Destination Port field of BTP-A Header
    source_port : int
        (16 bit integer) Source Port field of BTP-A Header
    """

    def __init__(self) -> None:
        self.destination_port = 0
        self.source_port = 0

    def initialize_with_request(self, request: BTPDataRequest) -> None:
        """
        Initializes the BTP-A Header with a GNDataRequest.

        Parameters
        ----------
        request : GNDataRequest
            GNDataRequest to use.
        """
        self.destination_port = request.destination_port
        self.source_port = request.source_port

    def encode_to_int(self) -> int:
        """
        Encodes the BTP-A Header to an integer.

        Returns
        -------
        int
            Encoded BTP-A Header
        """
        return (self.destination_port << 16) | self.source_port

    def encode(self) -> bytes:
        """
        Encodes the BTP-A Header to bytes.

        Returns
        -------
        bytes
            Encoded BTP-A Header
        """
        return self.encode_to_int().to_bytes(4, byteorder='big')

    def decode(self, data: bytes) -> None:
        """
        Decodes the BTP-A Header from bytes.

        Parameters
        ----------
        data : bytes
            Bytes to decode.
        """
        self.destination_port = int.from_bytes(data[0:2], byteorder='big')
        self.source_port = int.from_bytes(data[2:4], byteorder='big')


class BTPBHeader:
    """
    BTP-B Header class.

    Specified in ETSI EN 302 636-5-1 V2.1.0 (2017-05). Section 7.3

    Attributes
    ----------
    destination_port : int
        (16 bit integer) Destination Port field of BTP-B Header
    destination_port_info : int
        (16 bit integer) Destination Port Info field of BTP-B Header
    """

    def __init__(self) -> None:
        self.destination_port = 0
        self.destination_port_info = 0

    def initialize_with_request(self, request: BTPDataRequest) -> None:
        """
        Initializes the BTP-B Header with a GNDataRequest.

        Parameters
        ----------
        request : GNDataRequest
            GNDataRequest to use.
        """
        self.destination_port = request.destination_port
        self.destination_port_info = request.destinaion_port_info

    def encode_to_int(self) -> int:
        """
        Encodes the BTP-B Header to an integer.

        Returns
        -------
        int
            Encoded BTP-B Header
        """
        return (self.destination_port << 16) | self.destination_port_info

    def encode(self) -> bytes:
        """
        Encodes the BTP-B Header to bytes.

        Returns
        -------
        bytes
            Encoded BTP-B Header
        """
        return self.encode_to_int().to_bytes(4, byteorder='big')

    def decode(self, data: bytes) -> None:
        """
        Decodes the BTP-B Header from bytes.

        Parameters
        ----------
        data : bytes
            Bytes to decode.
        """
        self.destination_port = int.from_bytes(data[0:2], byteorder='big')
        self.destination_port_info = int.from_bytes(data[2:4], byteorder='big')



class InterfaceLDM1:
    """
    Class specified is described in TS 102 723-5.
    The IF.LDM.1 interface is responsible communication with the management layer.

    Attributes
    ----------
    TODO: Add attributes
    TODO: Finish once Managament Layer is implemented

    """
    def __init__(self) -> None:
        pass



import asn1tools

from .vam_asn1 import VAM_ASN1_DESCRIPTIONS


class VAMCoder:
    """
    Class for encoding/decoding Vulnerable Road User Awareness Message (VAM)

    Attributes
    ----------
    asn_coder : asn1tools.Coder
        ASN.1 coder.
    """

    def __init__(self) -> None:
        """
        Initialize the VAM Coder.
        """
        self.asn_coder = asn1tools.compile_string(
            VAM_ASN1_DESCRIPTIONS, codec='uper')

    def encode(self, vam: dict) -> bytes:
        """
        Encodes a VAM message.

        Parameters
        ----------
        vam : dict
            VAM message.

        Returns
        -------
        bytes
            Encoded VAM message.
        """
        return self.asn_coder.encode('VAM', vam)

    def decode(self, vam: bytes) -> dict:
        """
        Decodes a VAM message.

        Parameters
        ----------
        vam : bytes
            Encoded VAM message.

        Returns
        -------
        dict
            VAM message.
        """
        return self.asn_coder.decode('VAM', vam)



from enum import Enum

from .exceptions import DecodeError
from .mib import MIB


class BasicNH(Enum):
    """
    Next Header field class. As specified in ETSI EN 302 636-4-1 V1.4.1 (2020-01). Section 9.6.3

    Attributes
    ----------
    ANY :
        Any next header.
    COMMON_HEADER :
        Common Header.
    SECURED_PACKET :
        Secured Packet.
    """

    ANY = 0
    COMMON_HEADER = 1
    SECURED_PACKET = 2


class LTbase(Enum):
    """
    Lifetime base class. As specified in ETSI EN 302 636-4-1 V1.4.1 (2020-01). Section 9.6.4

    Attributes
    ----------
    fifty_milliseconds :
        50 ms.
    one_second :
        1 s.
    ten_seconds :
        10 s.
    one_hundred_seconds :
        100 s.
    """

    FIFTY_MILLISECONDS = 0
    ONE_SECOND = 1
    TEN_SECONDS = 2
    ONE_HUNDRED_SECONDS = 3


class LT:
    """
    Lifetime class. As specified in ETSI EN 302 636-4-1 V1.4.1 (2020-01). Section 9.6.4

    The Lifetime (LT) field shall indicate the maximum tolerable time a packet may be buffered until it reaches
    its destination.
    The LT field shall be comprised of two sub-fields: a LTMultiplier sub-field (Multiplier)
    and a LTBase sub-field (Base) (figure 10) and shall be encoded as follows:
    LT = Multiplier * Base

    Attributes
    ----------
    multiplier : int
        (5 bit unsigned integer) Lifetime multiplier.
    base : LTbase
        (2 bit unsigned integer) Lifetime base.
    """

    def __init__(self) -> None:
        self.multiplier = 0
        self.base = LTbase.FIFTY_MILLISECONDS

    def set_value_in_millis(self, value: int) -> None:
        """
        Set the lifetime in milliseconds.

        Parameters
        ----------
        value : int
            Lifetime in milliseconds.
        """
        if value < 50:
            self.multiplier = 0
            self.base = LTbase.FIFTY_MILLISECONDS
        elif value < 100:
            self.multiplier = 1
            self.base = LTbase.FIFTY_MILLISECONDS
        elif value < 500:
            self.multiplier = int(value / 50 % 64)
            self.base = LTbase.FIFTY_MILLISECONDS
        elif value < 1000:
            self.multiplier = 0
            self.base = LTbase.ONE_SECOND
        elif value < 10000:
            self.multiplier = int(value / 1000 % 64)
            self.base = LTbase.ONE_SECOND
        elif value < 100000:
            self.multiplier = int(value / 10000 % 64)
            self.base = LTbase.TEN_SECONDS
        elif value < 1000000:
            self.multiplier = int(value / 100000 % 64)
            self.base = LTbase.ONE_HUNDRED_SECONDS
        else:
            self.multiplier = 0
            self.base = LTbase.ONE_HUNDRED_SECONDS

    def set_value_in_seconds(self, value: int) -> None:
        """
        Set the lifetime in seconds.

        Parameters
        ----------
        value : int
            Lifetime in seconds.
        """
        self.set_value_in_millis(value * 1000)

    def get_value_in_millis(self) -> int:
        """
        Get the lifetime in milliseconds.

        Returns
        -------
        int
            Lifetime in milliseconds.
        """
        if self.base == LTbase.FIFTY_MILLISECONDS:
            return self.multiplier * 50
        if self.base == LTbase.ONE_SECOND:
            return self.multiplier * 1000
        if self.base == LTbase.TEN_SECONDS:
            return self.multiplier * 10000
        if self.base == LTbase.ONE_HUNDRED_SECONDS:
            return self.multiplier * 100000
        return 0

    def get_value_in_seconds(self) -> int:
        """
        Get the lifetime in seconds.

        Returns
        -------
        int
            Lifetime in seconds.
        """
        return self.get_value_in_millis() // 1000

    def encode_to_int(self) -> int:
        """
        Encode the LT to an integer.

        Returns
        -------
        int
            Encoded LT.
        """
        return self.multiplier << 2 | self.base.value

    def encode_to_bytes(self) -> bytes:
        """
        Encode the LT to bytes.

        Returns
        -------
        bytes
            Encoded LT.
        """
        return self.encode_to_int().to_bytes(1, "big")


class BasicHeader:
    """
    Basic Header class. As specified in ETSI EN 302 636-4-1 V1.4.1 (2020-01). Section 9.6

    Attributes
    ----------
    version : int
        (4 bit unsigned intger) Version of the GeoNetworking protocol.
    nh : int
        (4 bit unsigned integer) Next Header. Indicates the type of the next header.
    reserved : int
        (8 bit unsigned integer) Reserved. All bits set to zero.
    lt : LT
        (8 bit unsigned integer) Lifetime. Indicates the lifetime of the packet.
    rhl : int
        (8 bit unsigned integer) Remaining Hop Limit. Indicates the remaining number of hops.

    """

    def __init__(self) -> None:
        self.version = 1
        self.nh = BasicNH.COMMON_HEADER
        self.reserved = 0
        self.lt = LT()
        self.rhl = 0

    def set_version(self, version: int) -> None:
        """
        Set the version.

        Parameters
        ----------
        version : int
            Version of the GeoNetworking protocol.
        """
        self.version = version

    def set_nh(self, nh: BasicNH) -> None:
        """
        Set the next header.

        Parameters
        ----------
        nh : BasicNH
            Next Header. Indicates the type of the next header.
        """
        self.nh = nh

    def set_lt(self, lt: LT) -> None:
        """
        Set the lifetime.

        Parameters
        ----------
        lt : LT
            Lifetime. Indicates the lifetime of the packet.
        """
        self.lt = lt

    def set_rhl(self, rhl: int) -> None:
        """
        Set the remaining hop limit.

        Parameters
        ----------
        rhl : int
            Remaining Hop Limit. Indicates the remaining number of hops.
        """
        self.rhl = rhl % 256

    def encode_to_int(self) -> int:
        """
        Encode the Basic Header to an integer.

        Returns
        -------
        int
            Encoded Basic Header.
        """
        return (
            self.version << (4 + 3 * 8)
            | self.nh.value << (0 + 3 * 8)
            | self.reserved << 2 * 8
            | self.lt.encode_to_int() << 8
            | self.rhl
        )

    def encode_to_bytes(self) -> bytes:
        """
        Encode the Basic Header to bytes.

        Returns
        -------
        bytes
            Encoded Basic Header.
        """
        return self.encode_to_int().to_bytes(4, "big")

    def decode_from_int(self, value: int) -> None:
        """
        Decode the Basic Header from an integer.

        Parameters
        ----------
        value : int
            Encoded Basic Header.
        """
        self.version = value >> (4 + 3 * 8) & 0xF
        self.nh = BasicNH((value >> (0 + 3 * 8) & 0xF))
        self.reserved = value >> 2 * 8 & 0xFF
        self.lt.multiplier = (value >> 10) & 0x3F
        self.lt.base = LTbase((value >> 8) & 0x03)
        self.rhl = value & 0xFF

    def decode_from_bytes(self, value: bytes) -> None:
        """
        Decode the Basic Header from bytes.

        Parameters
        ----------
        value : bytes
            Encoded Basic Header.
        """
        if len(value) < 4:
            raise DecodeError("Basic Header must be 4 bytes long")
        self.decode_from_int(int.from_bytes(value[0:4], "big"))

    def initialize_with_mib(self, mib: MIB) -> None:
        """
        Initialize the Basic Header with the MIB.

        Parameters
        ----------
        mib : MIB
            MIB.
        """
        self.set_version(mib.itsGnProtocolVersion)
        self.set_nh(BasicNH.COMMON_HEADER)
        self.set_rhl(mib.itsGnDefaultHopLimit)
        self.reserved = 0
        self.lt.set_value_in_seconds(mib.itsGnDefaultPacketLifetime)

    def __eq__(self, __value: object) -> bool:
        """
        Check if two Basic Headers are equal.

        Parameters
        ----------
        __value : object
            Basic Header to compare to.
        """
        if not isinstance(__value, BasicHeader):
            return NotImplemented
        return (
            self.version == __value.version
            and self.nh == __value.nh
            and self.reserved == __value.reserved
            and self.lt == __value.lt
            and self.rhl == __value.rhl
        )



# pylint: skip-file
from flexstack.utils.asn1.etsi_its_cdd import ETSI_ITS_CDD_ASN1_DESCRIPTIONS

DENM_ASN1_DESCRIPTIONS = ETSI_ITS_CDD_ASN1_DESCRIPTIONS+\
    """
--! @options: no-fields-header

DENM-PDU-Description {itu-t (0) identified-organization (4) etsi (0) itsDomain (5) wg1 (1) denmPduRelease2 (103831) 
                      major-version-2 (2) minor-version-2 (2)} 

DEFINITIONS AUTOMATIC TAGS ::=

BEGIN

IMPORTS 
 
ActionId, ActionIdList, CauseCodeV2, ClosedLanes, DangerousGoodsExtended, DeltaReferencePosition, DeltaTimeMilliSecondPositive, DeltaTimeSecond, 
EnergyStorageType, EventZone, GeneralizedLanePositions, HeightLonCarr, InformationQuality, ItsPduHeader, IvimReferences, LanePosition, 
LightBarSirenInUse, ItineraryPath, MapReferences, MetaInformation, NumberOfOccupants, ObjectFace, OccupiedLanesWithConfidence, PerceivedObject, 
Position1d, PosLonCarr, PosCentMass, PositionOfPillars, PositioningSolutionType, PosFrontAx, PositionOfOccupants, PathPredictedList, 
RequestResponseIndication, ReferencePosition, RestrictedTypes, RoadConfigurationSectionList, RoadType, StandardLength3b,StandardLength12b, 
StationType, Speed, SpeedLimit, StationId, StationarySince, TimestampIts, Traces, TracesExtended, TrafficDirection, TrafficRule, TurningRadius, 
Temperature, VehicleMass, VehicleIdentification, Wgs84Angle, WheelBaseVehicle 
   
FROM ETSI-ITS-CDD {itu-t (0) identified-organization (4) etsi (0) itsDomain (5) wg1 (1) ts (102894) cdd (2) 
                   major-version-4 (4) minor-version-2 (2)};

/** 
 * This type represents the DENM PDU.
 *
 * It shall include the following components: 
 *
 * @field header: the header of the DENM PDU.
 *
 * @field denm: the payload of the DENM PDU.
*/
DENM ::= SEQUENCE {
	header ItsPduHeader (WITH COMPONENTS {... , protocolVersion (2), messageId(denm)}),
	denm   DenmPayload
}

/** 
 * This type represents the DENM payload.
 *
 * It shall include the following components: 
 *
 * @field management: the Management Container.
 *
 * @field situation: the optional Situation Container.
 *
 * @field location: the optional Location Container.
 *
 * @field alacarte: the optional ALaCarte Container .
 *
*/
DenmPayload ::= SEQUENCE {
    management ManagementContainer,
    situation  SituationContainer OPTIONAL,
    location   LocationContainer OPTIONAL,
    alacarte   AlacarteContainer OPTIONAL
}
((WITH COMPONENTS {..., management (WITH COMPONENTS {..., termination ABSENT}), situation PRESENT, location PRESENT}) |
 (WITH COMPONENTS {..., management (WITH COMPONENTS {..., termination PRESENT}), situation ABSENT, location ABSENT, alacarte ABSENT}))

/** 
 * This type represents the Management Container.
 *
 * It shall include the following components: 
 *
 * @field actionId: the identifier of the DENM.
 *
 * @field detectionTime: the time at which the event is detected.
 *
 * @field referenceTime: the time at which a new DENM, an update DENM or a cancellation DENM is generated
 *
 * @field termination: the optional termination type of the DENM.
 *
 * @field eventPosition: the geographical position used in the definition of the awareness area / relevance zone, see clause 6.1.3.
 *
 * @field awarenessDistance: the optional radius of the circular awareness area, with centre at the event position or at any of the 
 * event history points as defined in clause 6.1.3.1.
 *
 * @field trafficDirection: the optional traffic direction along which the receiving ITS-S may encounter the event, 
 * as defined in clause 6.1.3.
 *
 * @field validityDuration: the validity duration of a DENM. This component represents a time offset in the unit of second since detectionTime.
 *
 * @field transmissionInterval: the optional time interval for DENM transmission as defined by the originating ITS-S. 
 * If the component is not present in DENM, a default value defaultValidity is assumed.
 *
 * @field stationType: the station type information of the originating ITS-S.
 *
*/
ManagementContainer ::= SEQUENCE {
    actionId                  ActionId,
    detectionTime             TimestampIts,
    referenceTime             TimestampIts,
    termination               Termination OPTIONAL,
    eventPosition             ReferencePosition,
    awarenessDistance         StandardLength3b OPTIONAL,
    trafficDirection          TrafficDirection OPTIONAL,
    validityDuration          DeltaTimeSecond DEFAULT defaultValidity, 
    transmissionInterval      DeltaTimeMilliSecondPositive OPTIONAL,
    stationType               StationType,
    ...
}

/** 
 * This type represents the situation container.
 *
 * It shall include the following components: 
 *
 * @field informationQuality: the quality level of the information provided by the ITS-S application of the originating ITS-S. 
 * It indicates the probability of the detected event being truly existent at the event position.
 *
 * @field eventType: the event type, including direct cause and sub cause.
 *
 * @field linkedCause: the optional type of a linked event co-existing at the same time and the same place (same event zone), 
 * including direct cause and sub cause of the linked event, for which no other DENM is sent out.
 *
 * @field eventZone: an optional list of EventPoint, using the position indicated in the component eventPosition of the Management container 
 * as the reference position for the first EventPoint.
 *
 * @field linkedDenms: the optional list of DF ActionId, pointing to DENMs that are semantically connected because applying to consecutive 
 * event zones at the same time.
 *
 * @field eventEnd: the end position of the event along the road that is affected by the event w.r.t. the component eventPosition of the 
 * Management container. This end position is represented by the length of the event along the road. 
 *
*/
SituationContainer ::= SEQUENCE {
    informationQuality InformationQuality,
    eventType          CauseCodeV2,
    linkedCause        CauseCodeV2 OPTIONAL,
    eventZone          EventZone OPTIONAL,
    ...,
[[  linkedDenms        ActionIdList OPTIONAL,
    eventEnd           Position1d OPTIONAL ]]   
}
  ((WITH COMPONENTS {..., eventZone PRESENT, eventEnd ABSENT}) |
   (WITH COMPONENTS {..., eventZone ABSENT, eventEnd PRESENT}) |
   (WITH COMPONENTS {..., eventZone ABSENT, eventEnd ABSENT})) 
 


/** 
 * This type represents the Location Container.
 *
 * It shall include the following components: 
 *
 * @field eventSpeed: optional speed of a detected dynamic event and the confidence of the speed information. 
 *
 * @field eventPositionHeading: the optional heading of a dynamic event and the confidence of the heading information.
 *
 * @field detectionZonesToEventPosition: the detection zone information approaching the event position, see clause 6.1.3.3.
 *
 * @field roadType: the optional road type information at the event position. 
 *
 * @field lanePositions: the optional lane(s) where the event is located, at the position indicated by the component eventPosition 
 * of the Management container and for a given reference direction.
 *
 * @field occupiedLanes: the optional lane(s) that are fully or partially occupied by the event, at the position indicated by the 
 * component eventPosition of the Management container and for a given reference direction.
 *
 * @field linkedIvims: the optional list of DF IvimReference, pointing to IVIMs that are semantically connected because providing information 
 * applying to the road segment(s) covered by the components detectionZonesToEventPosition, detectionZonesToSpecifiedEventPoint and 
 * the SituationContainer component eventZone.
 *
 * @field linkedMapem: the optional list of DF Mapreference, pointing to MAPEMs that are semantically connected because providing information 
 * applying to the road segment(s) covered by the component detectionZonesToEventPosition, detectionZonesToSpecifiedEventPoint and 
 * the SituationContainer component eventZone.
 *
 * @field detectionZonesToSpecifiedEventPoint: the optional detection zone information approaching towards a 
 * specified event point, see clause 6.1.3.3. 
 *
 * @field predictedPaths: the optional list of future paths or trajectories that the event may move along or zones that the event may occupy. 
 *
*/
LocationContainer ::= SEQUENCE {
    eventSpeed                            Speed OPTIONAL,
    eventPositionHeading                  Wgs84Angle OPTIONAL,
    detectionZonesToEventPosition         Traces,
    roadType                              RoadType OPTIONAL,
    ...,
[[	lanePositions                         GeneralizedLanePositions OPTIONAL,
    occupiedLanes                         OccupiedLanesWithConfidence OPTIONAL,
    linkedIvims                           IvimReferences OPTIONAL, 
    linkedMapems                          MapReferences OPTIONAL, 
    detectionZonesToSpecifiedEventPoint   TracesExtended OPTIONAL,
    predictedPaths	          	          PathPredictedList OPTIONAL ]]
}

/** 
 * This type contains detailed information about the vehicle in which the originating ITS-S is mounted, for mitigating the consequences 
 * of a collision.
 *
 * It shall include the following components: 
 *
 * @field heightLonCarrLeft: the height of the left longitudinal carrier of the vehicle from base to top.
 *
 * @field heightLonCarrRight: the height of the right longitudinal carrier of the vehicle from base to top.
 *
 * @field posLonCarrLeft: the position of the left longitudinal carrier of vehicle.
 *
 * @field posLonCarrRight: the position of the right longitudinal carrier of vehicle.
 *
 * @field positionOfPillars: information about the vertical support of the vehicle in which the originating ITS-S is mounted. It shall be 
 * included for passenger vehicles only.
 *
 * @field posCentMass: the position of the centre of mass of the vehicle.
 *
 * @field wheelBaseVehicle: the wheel base of the vehicle.
 *
 * @field turningRadius: the turning radius of the vehicle.
 *
 * @field posFrontAx: the position of the front axle of the vehicle.
 *
 * @field positionOfOccupants: indicates whether an in-vehicle seat is occupied at the moment of generation of the message.
 *
 * @field vehicleMass: the mass of the unloaded vehicle
 *
 * @field requestResponseIndication: indicates whether the originating ITS-S transmitting the impactReduction component is requesting 
 * the receiving ITS-S to provide also its impactReduction component.
 *
*/
ImpactReductionContainer ::= SEQUENCE {
    heightLonCarrLeft         HeightLonCarr,
    heightLonCarrRight        HeightLonCarr,
    posLonCarrLeft            PosLonCarr,
    posLonCarrRight           PosLonCarr,
    positionOfPillars         PositionOfPillars,
    posCentMass               PosCentMass,
    wheelBaseVehicle          WheelBaseVehicle,
    turningRadius             TurningRadius,
    posFrontAx                PosFrontAx,
    positionOfOccupants       PositionOfOccupants,
    vehicleMass               VehicleMass,
    requestResponseIndication RequestResponseIndication
}

/** 
 * This type contains detailed information about an object with which a vehicle and/or the traffic is likely to collide.
 *
 * It shall include the following components: 
 *
 * @field perceivedPreCrashObject: information about a perceived object in the East, North, Up reference frame.
 *
 * @field objectStationId: the optional station Id of the object for which the information is provided.
 *
 * @field timeToCollision: the optional estimated time to collision of a vehicle with the object. 
 *
 * @field impactSection: indication of the object's section where the impact will most likely occur. 
 * When the target object is likely to be a vehicle, then this component should be present, otherwise it should not be provided.  
 *
 * @field estimatedBrakingDistance: the optional estimated distance in which the vehicle would need to come to a complete hold, 
 * if no obstruction was in the way. 
*/
PreCrashContainer ::= SEQUENCE { 
    perceivedPreCrashObject   PerceivedObject, 
    objectStationId           StationId OPTIONAL, 
    timeToCollision           DeltaTimeMilliSecondPositive OPTIONAL, 
    impactSection             ObjectFace OPTIONAL, 
    estimatedBrakingDistance  StandardLength12b OPTIONAL,
    ... 
}

/** 
 * This type contains detailed information about the configuration of road section(s) that are geographically related to the event.
 *
 * It shall include the following components: 
 *
 * @field roadConfigurationConfidence: information about the source of the road configuration and the confidence in the information.
 *
 * @field roadConfigurationSectionList: a list of road configuration information per applicable road section. 
*
*/
RoadConfigurationContainer ::=  SEQUENCE {
    roadConfigurationConfidence     MetaInformation,
    roadConfigurationSectionList    RoadConfigurationSectionList,
    ...
} 

/** 
 * This type contains detailed information of a roadwork zone and specific access conditions.
 *
 * It shall include the following components: 
 *
 * @field lightBarSirenInUse: optionally indicates whether a roadwork vehicle has switched on the light bar or siren. 
 * It is used when the roadwork involves a specific roadwork vehicle
 *
 * @field closedLanes: optionally indicates whether the roadwork has caused the closure of one or several driving lanes. 
 * Optionally, it may indicate whether a hard shoulder lane is closed to traffic or can be used for specific usage (e.g. for stopping).
 *
 * @field restriction: the optional type(s) of vehicles that are restricted to access the road work zone. 
 * More than one vehicle types may be provided by this component if the restriction apply to multiple vehicle types. 
 *
 * @field speedLimit: the optional speed limitation applied to the roadwork zone.
 *
 * @field incidentIndication: the optional incident related to the roadworks to provide additional information of the roadworks zone.
 *
 * @field recommendedPath: the optional recommended itinerary in order to contour the roadworks zone. If present, a recommended path 
 * shall be a list of path points in the order from the starting point closest to the roadworks zone to the end point of the recommended path. 
 *
 * @field startingPointSpeedLimit: the optional effective starting position of a speed limit being applied to the roadwork zone, 
 * with respect to the component eventPosition on the Management Container.
 * This component shall be present if the speed limit is applied at a certain distance prior to the roadwork zone starting position.
 *
 * @field trafficFlowRule: optionally indicates the side of the road to which the traffic should flow around a roadwork.
 *
 * @field referenceDenms: an optional sequence of actionIds for different DENMs that describe the same event. 
 * If it is available, it indicates the actionIds of all other DENMs describing this event.
 *
*/
RoadWorksContainerExtended ::= SEQUENCE {
    lightBarSirenInUse      LightBarSirenInUse OPTIONAL,
    closedLanes             ClosedLanes OPTIONAL,
    restriction             RestrictedTypes OPTIONAL,
    speedLimit              SpeedLimit OPTIONAL,
    incidentIndication      CauseCodeV2 OPTIONAL,
    recommendedPath         ItineraryPath OPTIONAL,
    startingPointSpeedLimit DeltaReferencePosition OPTIONAL,
    trafficFlowRule         TrafficRule OPTIONAL,
    referenceDenms          ActionIdList OPTIONAL
}

/** 
 * This type contains detailed information about a stationary vehicle.
 *
 * It shall include the following components: 
 *
 * @field stationarySince: the optional time duration of the stationary vehicle being stationary.
 *
 * @field stationaryCause: optional additional information to describe causes of the stationary vehicle event such as human problem.
 *
 * @field carryingDangerousGoods: optional information on the type of dangerous goods, the required emergency action and other information.
 *
 * @field numberOfOccupants: the optional estimated number of occupants involved in the stationary vehicle event.
 *
 * @field vehicleIdentification: the optional identification of the stationary vehicle.
 *
 * @field energyStorageType: the optional vehicle energy storage type information of the stationary vehicle, such as electric, diesel, etc.
 *
*/
StationaryVehicleContainer ::= SEQUENCE {
    stationarySince        StationarySince OPTIONAL, 
    stationaryCause        CauseCodeV2 OPTIONAL,
    carryingDangerousGoods DangerousGoodsExtended OPTIONAL,
    numberOfOccupants      NumberOfOccupants OPTIONAL,
    vehicleIdentification  VehicleIdentification OPTIONAL,
    energyStorageType      EnergyStorageType OPTIONAL
}

/** 
 * This type represents the A La Carte Container.
 *
 * It shall include the following components: 
 *
 * @field lanePosition: the optional lane position of the event.
 *
 * @field impactReduction: optional vehicle data for collision mitigation.
 *
 * @field externalTemperature: optional the ambient temperature at the event position.
 *
 * @field roadWorks: optional information of the roadwork zone.
 *
 * @field positioningSolution: optionally indicates the technical solution being used by the originating ITS-S to estimate the event position.
 *
 * @field stationaryVehicle: optional information about a stationary vehicle.
 *
 * @field roadConfiguration: optional information about the configuration of the road.
 *
 * @field precrash: the optional information about perceived objects that represent hazards and/or could be subject of collisions. 
 *
*/
AlacarteContainer ::= SEQUENCE {
    lanePosition        LanePosition OPTIONAL,
    impactReduction     ImpactReductionContainer OPTIONAL,
    externalTemperature Temperature OPTIONAL,
    roadWorks           RoadWorksContainerExtended OPTIONAL,
    positioningSolution PositioningSolutionType OPTIONAL,
    stationaryVehicle   StationaryVehicleContainer OPTIONAL,
    ...,
[[  roadConfiguration 	RoadConfigurationContainer OPTIONAL,
    preCrash            PreCrashContainer OPTIONAL ]]
}

/** 
 * This type specifies the default value for DENM validity duration used for DENM protocol operation.
*/
defaultValidity INTEGER ::= 600

/**
  * This indicates the termination type of generated DENM, i.e. if it is a cancellation DENM or a negation DENM
*/

Termination ::= ENUMERATED {isCancellation(0), isNegation (1)}

END

"""


from __future__ import annotations
import time
import json
import logging

from .ldm_classes import Location, RequestDataObjectsReq, TimestampIts
from .ldm_constants import (
    MAINTENANCE_AREA_MAX_ALTITUDE_DIFFERENCE,
    NEW_DATA_RECIEVED,
    NO_NEW_DATA_RECIEVED,
)
from .database import DataBase
from .ldm_classes import Utils


class LDMMaintenance:
    """
    Class specified in the ETSI EN 302 895 V1.1.1 (2014-09). Section 5.3.2
    The LDM Maintenance component (see clause 5.3.2) is responsible for storing and maintaining the
    data and its integrity as well as for the garbage collection of persistent data held within the LDM.

    Attributes
    ----------
    TODO: Add attributes
    """

    def __init__(self, area_of_maintenance: Location, data_base: DataBase) -> None:
        self.logging = logging.getLogger("local_dynamic_map")

        self.data_containers = data_base
        self.area_of_maintenance = area_of_maintenance
        self.new_data_recieved_flag = NO_NEW_DATA_RECIEVED

    def run(self) -> None:
        """
        Method specified in the ETSI 302 895 V1.1.1 (2014-09). Section 5.3.2.
        This method is called when the thread is started and has to mantain the data containers
        according to the timestamp, location and area of maintenance.

        Parameters
        ----------
        None
        """
        self.logging.info("LDMMaintenance trash collection initiated")

        while True:
            time.sleep(1)
            self.collect_trash()

    def delete_all_database(self) -> None:
        """
        Method created in order to delete all the data containers.

        Parameters
        ----------
        None
        """
        self.data_containers.delete()

    def add_provider_data(self, data: dict) -> int:
        """
        Method created in order to add data into the data containers.

        Parameters
        ----------
        data : dict
        """
        try:
            doc_id = self.data_containers.insert(dict(data))
            self.new_data_recieved_flag = NEW_DATA_RECIEVED
        except (KeyError, json.decoder.JSONDecodeError) as e:
            print(f"Error adding data container: {str(e)}")
            doc_id = None

        return doc_id

    def get_provider_data(self, data_object_id) -> dict:
        """
        Method created in order to get data from the data containers.

        Parameters
        ----------
        data_object_id : int
        """
        try:
            provider_data = self.data_containers.get(doc_id=data_object_id)
        except (KeyError, json.decoder.JSONDecodeError) as e:
            print(f"Error getting data container: {str(e)}")
            provider_data = None

        return provider_data

    def update_provider_data(self, data_object_id, data_object) -> None:
        """
        Method created in order to update data from the data containers.

        Parameters
        ----------
        data_object_id : int
        data_object : dict
        """
        try:
            self.data_containers.update(
                data_object,
                doc_ids=[
                    data_object_id,
                ],
            )
            self.logging.debug("Data container updated: %s", data_object_id)
        except (KeyError, json.decoder.JSONDecodeError) as e:
            print(f"Error updating data container: {str(e)}")

    def del_provider_data(self, data_object: dict = None) -> None:
        """
        Method created in order to delete data from the data containers.

        Parameters
        ----------
        id : int
        """
        try:
            self.data_containers.remove(data_object)
        except (KeyError, json.decoder.JSONDecodeError) as e:
            print(f"Error deleting data container: {str(e)}, data_containers {len(self.data_containers.all())}")

    def get_all_data_containers(self) -> list[dict]:
        """
        Method created in order to get all the data containers.

        Parameters
        ----------
        None
        """
        try:
            data_containers = self.data_containers.all()
        except (KeyError, json.decoder.JSONDecodeError) as e:
            print(f"Error getting all data containers: {str(e)}")
            data_containers = []

        return data_containers

    def search_data_containers(self, data_request: RequestDataObjectsReq) -> list[dict]:
        """
        Method created in order to search data from the data containers.
        It redirects the search of the data containers to the database.

        Parameters
        ----------
        filter : Filter

        Returns
        -------
        search_result : list[dict]
        """

        search_result = self.data_containers.search(data_request)

        return search_result

    def check_and_delete_time_validity(self) -> list[int]:
        """
        Method according to ETSI EN 302 895 V1.1.1 (2014-09). Section 5.3.2

        Method checks if the data is still valid according to the time validity period.

        Parameters
        ----------
        None
        """

        time_invalidity_data_containers = []
        try:
            for data_container in self.get_all_data_containers():
                if (data_container["timeValidity"] + data_container["timeStamp"]) < TimestampIts(
                    time.time()
                ).convert_epoch_to_its_timestamp():
                    self.del_provider_data(data_container)
                    time_invalidity_data_containers.append(data_container)
        except json.decoder.JSONDecodeError as e:
            print(f"Error checking time validity: {str(e)}")
        return time_invalidity_data_containers

    def check_and_delete_area_of_maintenance(self) -> list[int]:
        """
        Method according to ETSI EN 302 895 V1.1.1 (2014-09). Section 5.3.2

        Method checks if the data is still valid according to the area of maintenance.

        Parameters
        ----------
        None
        """
        area_maintenance_invalidity_data_containers = []
        for data_container in self.get_all_data_containers():
            distance_between_points_and_maintenance_area = Utils.euclidian_distance(
                (
                    data_container["location"]["referencePosition"]["latitude"],
                    data_container["location"]["referencePosition"]["longitude"],
                ),
                (
                    self.area_of_maintenance.reference_position.latitude,
                    self.area_of_maintenance.reference_position.longitude,
                ),
            )
            if (
                self.area_of_maintenance.reference_area.relevance_area.relevance_distance.compare_with_int(
                    distance_between_points_and_maintenance_area
                )
                and (
                    data_container["location"]["referencePosition"]["altitude"]["altitudeValue"]
                    - self.area_of_maintenance.reference_position.altitude.altitude_value
                )
                ^ 2
                < MAINTENANCE_AREA_MAX_ALTITUDE_DIFFERENCE
            ):
                self.del_provider_data(data_container)
                area_maintenance_invalidity_data_containers.append(data_container)

        return area_maintenance_invalidity_data_containers

    def collect_trash(self) -> None:
        """
        Method according to ETSI EN 302 895 V1.1.1 (2014-09). Section 5.3.2

        Method should be called periodically to delete expired data from the data containers.
        Expired data consists of:
        - Data that doesn't meet its time validity (time validity period): Clause 6.2.3
        - Data that is not within the LDM Area of Maintance

        Parameters
        ----------
        None
        """
        initial_data_containers = len(self.get_all_data_containers())
        self.check_and_delete_time_validity()
        # Check data is within the LDM Area of Maintenance
        self.check_and_delete_area_of_maintenance()

        self.logging.debug(
            "Deleted %s/%s data containers in trash collection",
            initial_data_containers - len(self.get_all_data_containers()),
            initial_data_containers
        )

    def update_area_of_maintenance(self, area_of_maintenance) -> None:
        """
        Method according to ETSI EN 302 895 V1.1.1 (2014-09). Section 5.3.2

        Method should be called periodically to update the area of maintance.

        Parameters
        ----------
        area_of_maintenance : list[int, int, int]
        """
        self.area_of_maintenance = area_of_maintenance

    def check_new_data_recieved(self) -> int:
        """
        Method used to verify if new data has been recieved. Uses threading locks to avoid any type of overwritting.

        Parameters
        ----------
        None
        """
        if self.new_data_recieved_flag == NEW_DATA_RECIEVED:
            self.new_data_recieved_flag = NO_NEW_DATA_RECIEVED
            return NEW_DATA_RECIEVED
        return NO_NEW_DATA_RECIEVED



# pylint: skip-file
ETSI_ITS_DSRC_REGION_ASN1_DESCRIPTIONS = """
--! @options: no-fields-header

ETSI-ITS-DSRC-REGION {
  itu-t (0) identified-organization (4) etsi (0) itsDomain (5) wg1 (1) ts103301 (103301) dsrc (6) region (1) major-version-2 (2) minor-version-1 (1)
}

DEFINITIONS AUTOMATIC TAGS::= BEGIN

IMPORTS

addGrpC, REG-EXT-ID-AND-TYPE
FROM ETSI-ITS-DSRC {
  itu-t (0) identified-organization (4) etsi (0) itsDomain (5) wg1 (1) ts103301 (103301) dsrc (6) major-version-2 (2) minor-version-1 (1)
}

ConnectionManeuverAssist-addGrpC, ConnectionTrajectory-addGrpC,
IntersectionState-addGrpC, LaneAttributes-addGrpC, MapData-addGrpC,
MovementEvent-addGrpC, NodeAttributeSet-addGrpC, Position3D-addGrpC, RequestorDescription-addGrpC, RestrictionUserType-addGrpC, SignalStatusPackage-addGrpC
FROM ETSI-ITS-DSRC-AddGrpC {
  itu-t (0) identified-organization (4) etsi (0) itsDomain (5) wg1 (1) ts103301 (103301) dsrc (6) addgrpc (0) major-version-2 (2) minor-version-1 (1)
};

Reg-AdvisorySpeed	            REG-EXT-ID-AND-TYPE ::= { ... }

Reg-ComputedLane	            REG-EXT-ID-AND-TYPE ::= { ... }

Reg-ConnectionManeuverAssist	REG-EXT-ID-AND-TYPE ::= {
	{ConnectionManeuverAssist-addGrpC  IDENTIFIED BY addGrpC},
	...
}

Reg-GenericLane	                REG-EXT-ID-AND-TYPE ::= {
	{ConnectionTrajectory-addGrpC	IDENTIFIED BY addGrpC} ,
	...
}

Reg-IntersectionGeometry  	    REG-EXT-ID-AND-TYPE ::= { ... }

Reg-IntersectionState           REG-EXT-ID-AND-TYPE ::= {
	{IntersectionState-addGrpC IDENTIFIED BY addGrpC},
	...
}

Reg-LaneAttributes	            REG-EXT-ID-AND-TYPE ::= {
   {LaneAttributes-addGrpC IDENTIFIED BY addGrpC} ,
   ...
}
Reg-LaneDataAttribute           REG-EXT-ID-AND-TYPE ::= { ... }

Reg-MapData	                    REG-EXT-ID-AND-TYPE ::= {
	{MapData-addGrpC  IDENTIFIED BY addGrpC},
	...
}

Reg-MovementEvent	            REG-EXT-ID-AND-TYPE ::= {
   {MovementEvent-addGrpC IDENTIFIED BY addGrpC} ,
   ...
}
Reg-MovementState               REG-EXT-ID-AND-TYPE ::= { ... }


Reg-NodeAttributeSetXY          REG-EXT-ID-AND-TYPE ::= {
	{NodeAttributeSet-addGrpC   IDENTIFIED BY addGrpC},
	...
}

Reg-NodeOffsetPointXY	        REG-EXT-ID-AND-TYPE ::= { ... }

Reg-Position3D	                REG-EXT-ID-AND-TYPE ::= {
	{Position3D-addGrpC  IDENTIFIED BY addGrpC} ,
	...
}

Reg-RequestorDescription        REG-EXT-ID-AND-TYPE ::= {
   { RequestorDescription-addGrpC IDENTIFIED BY addGrpC} ,
   ...
}

Reg-RequestorType	            REG-EXT-ID-AND-TYPE ::= { ... }

Reg-RestrictionUserType	        REG-EXT-ID-AND-TYPE ::= {
  {RestrictionUserType-addGrpC IDENTIFIED BY addGrpC} ,
  ...
}

Reg-RoadSegment	                REG-EXT-ID-AND-TYPE ::= { ... }

Reg-RTCMcorrections             REG-EXT-ID-AND-TYPE ::= { ... }

Reg-SignalControlZone           REG-EXT-ID-AND-TYPE ::= { ... }

Reg-SignalRequest               REG-EXT-ID-AND-TYPE ::= { ... }

Reg-SignalRequestMessage        REG-EXT-ID-AND-TYPE ::= { ... }

Reg-SignalRequestPackage        REG-EXT-ID-AND-TYPE ::= { ... }

Reg-SignalStatus	            REG-EXT-ID-AND-TYPE ::= { ... }

Reg-SignalStatusMessage	        REG-EXT-ID-AND-TYPE ::= { ... }

Reg-SignalStatusPackage	        REG-EXT-ID-AND-TYPE ::= {
	{ SignalStatusPackage-addGrpC IDENTIFIED BY addGrpC },
	...
}

Reg-SPAT	                    REG-EXT-ID-AND-TYPE ::= { ... }

END
    
"""
